#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define BLOCK_SIZE 16

__device__ float apply_leaky_relu(float x, float negative_slope) {
    return x > 0 ? x : x * negative_slope;
}

// Optimized CUDA kernel using shared memory tiling for GEMM followed by bias addition, scaling and LeakyReLU activation
__global__ void optimized_gemm_kernel(
    const float* __restrict__ x,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    const int batch_size,
    const int in_features,
    const int out_features,
    const float multiplier,
    const float negative_slope
) {
    __shared__ float s_x[BLOCK_SIZE][BLOCK_SIZE];
    __shared__ float s_w[BLOCK_SIZE][BLOCK_SIZE];

    const int tx = threadIdx.x;
    const int ty = threadIdx.y;
    const int row = blockIdx.x * BLOCK_SIZE + tx;
    const int col = blockIdx.y * BLOCK_SIZE + ty;
    float sum = 0.0f;

    const int numTiles = (in_features + BLOCK_SIZE - 1) / BLOCK_SIZE;
    const int row_base = row * in_features;
    const int col_base = col * in_features;

    for (int t = 0; t < numTiles; t++) {
        const int tile_idx = t * BLOCK_SIZE;
        
        // Load input tile - coalesced access
        if (row < batch_size && (tile_idx + ty) < in_features) {
            s_x[tx][ty] = x[row_base + tile_idx + ty];
        } else {
            s_x[tx][ty] = 0.0f;
        }

        // Load weight tile with transpose - coalesced access
        if (col < out_features && (tile_idx + tx) < in_features) {
            s_w[ty][tx] = weight[col_base + tile_idx + tx];
        } else {
            s_w[ty][tx] = 0.0f;
        }

        __syncthreads();

        #pragma unroll
        for (int k = 0; k < BLOCK_SIZE; k++) {
            sum += s_x[tx][k] * s_w[ty][k];
        }
        __syncthreads();
    }

    if (row < batch_size && col < out_features) {
        sum = (sum + bias[col]) * multiplier;
        output[row * out_features + col] = apply_leaky_relu(sum, negative_slope);
    }
}

// Host function that sets up kernel execution
torch::Tensor module_fn_forward(
    torch::Tensor x,
    float multiplier,
    float negative_slope,
    torch::Tensor weight,
    torch::Tensor bias
) {
    TORCH_CHECK(x.device().is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.device().is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.device().is_cuda(), "bias must be a CUDA tensor");

    const int batch_size = x.size(0);
    const int in_features = x.size(1);
    const int out_features = weight.size(0);

    TORCH_CHECK(weight.size(1) == in_features, "Weight in_features must match x in_features");
    TORCH_CHECK(bias.size(0) == out_features, "Bias size must match weight out_features");

    auto output = torch::zeros({batch_size, out_features}, x.options());

    dim3 block(BLOCK_SIZE, BLOCK_SIZE);
    dim3 grid((batch_size + BLOCK_SIZE - 1) / BLOCK_SIZE, (out_features + BLOCK_SIZE - 1) / BLOCK_SIZE);

    optimized_gemm_kernel<<<grid, block>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        output.data_ptr<float>(),
        batch_size,
        in_features,
        out_features,
        multiplier,
        negative_slope
    );

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn_forward, "Optimized module function forward CUDA with shared memory tiling and LeakyReLU");
}
